/*********************************************
 * vim:sw=8:ts=8:si:et
 * To use the above modeline in vim you must have "set modeline" in your .vimrc
 * Author: Guido Socher
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Ethernet remote device with LCD display
 * HTTP interface 
        url looks like this http://10.0.0.25/q&pw=secret&cmd=0
        or http://10.0.0.25/q&pw=secret&cmd=1 
        or http://10.0.0.25/q&pw=secret&cmd=2&l1=Hello&l2=You
 *
 * http://tuxgraphics.org/electronics/
 * Note: there is a version number in the text. Search for "tuxgraphics"
 *
 * ethlcd
 * (c) 2008 Mariusz Bialonczyk <manio@skyboo.net>
 * http://manio.skyboo.net/ethlcd/
 *
 *********************************************/
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <avr/pgmspace.h>
#include "ip_arp_udp_tcp.h"
#include "enc28j60.h"
#include "timeout.h"
#include "avr_compat.h"
#include "net.h"
#include "lcd.h"
#include "ethlcd_proto.h"

// please modify the following two lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices:
static uint8_t mymac[6] = {0x00,0x04,0x25,0x54,0x55,0x58};
// how did I get the mac addr? Translate the last 3 numbers into ascii is: TUX,
// first 3 bytes are Atmel company vendor ID
static uint8_t myip[4] = {10,1,0,7};
// base url (you can put a DNS name instead of an IP addr. if you have
// a DNS server (baseurl must end in "/"):
static char baseurl[]="http://10.1.0.7/";
static uint16_t mywwwport =80; // listen port for tcp/www
static uint16_t mydataport =2425; // listen port for tcp/dataport

#define FIRMWARE_VERSION "v0.1-20080813"
#define PROTOCOL_VERSION 1
#define WELCOME_LINE "ethlcd " FIRMWARE_VERSION

#define BUFFER_SIZE 750
static uint8_t buf[BUFFER_SIZE+1];
#define STR_BUFFER_SIZE 22
static char strbuf[STR_BUFFER_SIZE+1];

// the password string (only the first 5 char checked), (only a-z,0-9,_ characters):
static char password[]="secret"; // must not be longer than 9 char

// 
uint8_t verify_password(char *str)
{
        // a simple password/cookie:
        if (strcmp(password,str)==0){
                return(1);
        }
        return(0);
}

// search for a string of the form key=value in
// a string that looks like q?xyz=abc&uvw=defgh HTTP/1.1\r\n
//
// The returned value is stored in the global var strbuf
uint8_t find_key_val(char *str,char *key)
{
        uint8_t found=0;
        uint8_t i=0;
        char *kp;
        kp=key;
        while(*str &&  *str!=' ' && found==0){
                if (*str == *kp){
                        kp++;
                        if (*kp == '\0'){
                                str++;
                                kp=key;
                                if (*str == '='){
                                        found=1;
                                }
                        }
                }else{
                        kp=key;
                }
                str++;
        }
        if (found==1){
                // copy the value to a buffer and terminate it with '\0'
                while(*str &&  *str!=' ' && *str!='&' && i<STR_BUFFER_SIZE){
                        strbuf[i]=*str;
                        i++;
                        str++;
                }
                strbuf[i]='\0';
        }
        return(found);
}

// convert a single hex digit character to its integer value
unsigned char h2int(char c)
{
        if (c >= '0' && c <='9'){
                return((unsigned char)c - '0');
        }
        if (c >= 'a' && c <='f'){
                return((unsigned char)c - 'a' + 10);
        }
        if (c >= 'A' && c <='F'){
                return((unsigned char)c - 'A' + 10);
        }
        return(0);
}

// decode a url string e.g "hello%20joe" or "hello+joe" becomes "hello joe" 
void urldecode(char *urlbuf)
{
        char c;
        char *dst;
        dst=urlbuf;
        while ((c = *urlbuf)) {
                if (c == '+') c = ' ';
                if (c == '%') {
                        urlbuf++;
                        c = *urlbuf;
                        urlbuf++;
                        c = (h2int(c) << 4) | h2int(*urlbuf);
                }
                *dst = c;
                dst++;
                urlbuf++;
        }
        *dst = '\0';
}

// takes a string of the form q?pw=secret&cmd=1
// or q?pw=secret&cmd=2&l1=line+one&l2=two and analyse it
// The command number is 3 if the lcd data should be written.
// 
// return values: -1 invalid password, -2 no password
//                -3 no command given but password valid
//                any other number= command number
int8_t analyse_cmd(char *str)
{
        int8_t r=-3;
        if (find_key_val(str,"pw")){
                if (verify_password(strbuf)==0){
                        return(-1);
                }
        }else{
                // no pw found
                return(-2);
        }
        if (find_key_val(str,"cmd")){
                if (*strbuf < 0x3a && *strbuf > 0x2f){
                        // is a ASCII number, return it
                        r=(*strbuf-0x30);
                }
        }
        // cmd=1/0 is for the relay
        // cmd=2 is for the lcd display, read the l1 and l2 strings
        if (r==2){
                // read the lcd lines
                lcd_clrscr();
                if (find_key_val(str,"l1")){
                        urldecode(strbuf);
                        lcd_puts(strbuf);
                }
                if (find_key_val(str,"l2")){
                        lcd_gotoxy(0,1);
                        urldecode(strbuf);
                        lcd_puts(strbuf);
                }
        }
        return(r);
}

// prepare the login webpage by writing the data to the tcp send buffer
uint16_t print_loginpage(uint8_t *buf)
{
        uint16_t plen;
        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<center><form METHOD=\"get\" action=\""));
        plen=fill_tcp_data(buf,plen,baseurl);
        plen=fill_tcp_data_p(buf,plen,PSTR("q\">\npassword: <input size=8 type=\"password\" name=\"pw\"><input type=submit value=\"ok\"></form></center>"));
        return(plen);
}

// prepare the webpage by writing the data to the tcp send buffer
uint16_t print_webpage(uint8_t *buf,uint8_t on_off)
{
        uint16_t plen;
        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<center><p>Output is: "));
        if (on_off){
                plen=fill_tcp_data_p(buf,plen,PSTR("<font color=\"#00FF00\"> ON </font>"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("OFF"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR(" <a href=\""));
        plen=fill_tcp_data(buf,plen,baseurl);
        plen=fill_tcp_data_p(buf,plen,PSTR("q?pw="));
        plen=fill_tcp_data(buf,plen,password);
        plen=fill_tcp_data_p(buf,plen,PSTR("\">[refresh status]</a></p>\n<p><a href=\""));
        // the url looks like this http://baseurl/q?pw=secret&cmd=1
        plen=fill_tcp_data(buf,plen,baseurl);
        plen=fill_tcp_data_p(buf,plen,PSTR("q?pw="));
        plen=fill_tcp_data(buf,plen,password);
        if (on_off){
                plen=fill_tcp_data_p(buf,plen,PSTR("&cmd=0\">Switch off</a><p>"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("&cmd=1\">Switch on</a><p>"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("<hr><br><form METHOD=get action=\""));
        plen=fill_tcp_data(buf,plen,baseurl);
        plen=fill_tcp_data_p(buf,plen,PSTR("q\">\n<input size=20 type=text name=l1>\n<br><input size=20 type=text name=l2>\n<input type=hidden name=cmd value=2>\n<input type=hidden name=pw value=\""));
        plen=fill_tcp_data(buf,plen,password);
        plen=fill_tcp_data_p(buf,plen,PSTR("\">\n<br><input type=submit value=\"write to LCD\"></form>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("</center><hr><br>version 2.7, tuxgraphics.org\n"));
        return(plen);
}


void lcd_send_instr(uint8_t *data, uint16_t len)
{
	int16_t i=0;
	for (i=0; i<len; i++)
		lcd_command(data[i]);
}

void lcd_send_data(uint8_t *data, uint16_t len)
{
	int16_t i=0;
	for (i=0; i<len; i++)
		lcd_putc(data[i]);
}

int main(void){

        
        uint16_t plen;
        uint16_t dat_p;
        int8_t cmd;
        int16_t dlen;	//data length
        
        // set the clock speed to "no pre-scaler" (8MHz with internal osc or 
        // full external speed)
        // set the clock prescaler. First write CLKPCE to enable setting of clock the
        // next four instructions.
        CLKPR=(1<<CLKPCE); // change enable
        CLKPR=0; // "no pre-scaler"
        delay_ms(1);

        /* enable PD2/INT0, as input */
        DDRD&= ~(1<<DDD2);

        DDRC &= ~(1<<DDC1);	//C1 (LEFT button) as input
        PORTC |= 1<<PC1;
        DDRC &= ~(1<<DDC2);	//C2 (MENU button) as input
        PORTC |= 1<<PC2;
        DDRC &= ~(1<<DDC3);	//C3 (DOWN button) as input
        PORTC |= 1<<PC3;
        DDRC &= ~(1<<DDC4);	//C4 (ENTER button) as input
        PORTC |= 1<<PC4;
        DDRC &= ~(1<<DDC5);	//C5 (UP button) as input
        PORTC |= 1<<PC5;
        DDRB &= ~(1<<DDB1);	//B1 (RIGHT button) as input
        PORTB |= 1<<PB1;

        DDRC |= (1<<DDC0);	//C0 (beep) as output

        DDRB |= (1<<DDB0);	//PB0 (backlight) as output
        PORTB &= ~(1<<PB0);	//full light

        /*initialize enc28j60*/
        enc28j60Init(mymac);
        enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
        delay_ms(10);
        
        
        /* Magjack leds configuration, see enc28j60 datasheet, page 11 */
        // LEDB=yellow LEDA=green
        //
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);
        delay_ms(20);
        
        // lcd display:
        lcd_init(LCD_DISP_ON);
        lcd_clrscr();
        lcd_puts_P(WELCOME_LINE);
        //init the ethernet/ip layer:
        init_ip_arp_udp_tcp(mymac,myip,mywwwport);

        while(1){
                // get the next new packet:
                plen = enc28j60PacketReceive(BUFFER_SIZE, buf);

                /*plen will be unequal to zero if there is a valid 
                 * packet (without crc error) */
                if(plen==0){
                        continue;
                }
                        
                // arp is broadcast if unknown but a host may also
                // verify the mac address by sending it to 
                // a unicast address.
                if(eth_type_is_arp_and_my_ip(buf,plen)){
                        make_arp_answer_from_request(buf);
                        continue;
                }

                // check if ip packets are for us:
                if(eth_type_is_ip_and_my_ip(buf,plen)==0){
                        continue;
                }
                
                if(buf[IP_PROTO_P]==IP_PROTO_ICMP_V && buf[ICMP_TYPE_P]==ICMP_TYPE_ECHOREQUEST_V){
                        // a ping packet, let's send pong
                        make_echo_reply_from_request(buf,plen);
                        continue;
                }
                // tcp port www start
                if (buf[IP_PROTO_P]==IP_PROTO_TCP_V && buf[TCP_DST_PORT_H_P]==(mywwwport>>8) && buf[TCP_DST_PORT_L_P]==(mywwwport&0xFF)){
                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_SYN_V){
                                make_tcp_synack_from_syn(buf, mywwwport);
                                // make_tcp_synack_from_syn does already send the syn,ack
                                continue;
                        }
                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_ACK_V){
                                init_len_info(buf); // init some data structures
                                // we can possibly have no data, just ack:
                                dat_p=get_tcp_data_pointer();
                                if (dat_p==0){
                                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_FIN_V){
                                                // finack, answer with ack
                                                make_tcp_ack_from_any(buf, mywwwport);
                                        }
                                        // just an ack with no data, wait for next packet
                                        continue;
                                }
                                if (strncmp("GET ",(char *)&(buf[dat_p]),4)!=0){
                                        // head, post and other methods:
                                        //
                                        // for possible status codes see:
                                        // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                                        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<h1>200 OK</h1>"));
                                        goto SENDTCP;
                                }
                                if (strncmp("/ ",(char *)&(buf[dat_p+4]),2)==0){
                                        plen=print_loginpage(buf);
                                        goto SENDTCP;
                                }
                                cmd=analyse_cmd((char *)&(buf[dat_p+5]));
                                // for possible status codes see:
                                // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                                if (cmd==-1){
                                        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 401 Unauthorized\r\nContent-Type: text/html\r\n\r\n<h1>401 Unauthorized</h1>"));
                                        goto SENDTCP;
                                }
                                if (cmd==-2){
                                        plen=print_loginpage(buf);
                                        goto SENDTCP;
                                }
                                if (cmd==1){
                                        PORTB|= (1<<PB0);// transistor on
                                }
                                if (cmd==0){
                                        PORTB &= ~(1<<PB0);// transistor off
                                }
                                // just display the web-form:
                                plen=print_webpage(buf,(PORTB & (1<<PB0)));
                                //
SENDTCP:
                                make_tcp_ack_from_any(buf, mywwwport); // send ack for http get
                                make_tcp_ack_with_data(buf,plen,1); // send data
                                continue;
                        }

                }
                // tcp port www end

                // ethlcd tcp data port start
                if (buf[IP_PROTO_P]==IP_PROTO_TCP_V && buf[TCP_DST_PORT_H_P]==(mydataport>>8) && buf[TCP_DST_PORT_L_P]==(mydataport&0xFF)){
                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_SYN_V){
                                make_tcp_synack_from_syn(buf, mydataport);
                                // make_tcp_synack_from_syn does already send the syn,ack
                                continue;
                        }
                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_ACK_V){
                                dlen=init_len_info(buf); // init some data structures
                                // we can possibly have no data, just ack:
                                dat_p=get_tcp_data_pointer();
                                if (dat_p==0){
                                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_FIN_V){
                                                // finack, answer with ack
                                                make_tcp_ack_from_any(buf, mydataport);
                                        }
                                        // just an ack with no data, wait for next packet
                                        continue;
                                }
                                switch (buf[dat_p])	//first byte of data is command
				{
					case ETHLCD_SEND_INSTR:
						lcd_send_instr(&(buf[dat_p+1]), dlen-1);
						strbuf[0] = ETHLCD_SEND_INSTR;
						strbuf[1] = 0;
						plen=fill_tcp_data(buf,0,strbuf);
						break;
					case ETHLCD_SEND_DATA:
						lcd_send_data(&(buf[dat_p+1]), dlen-1);
						strbuf[0] = ETHLCD_SEND_DATA;
						strbuf[1] = 0;
						plen=fill_tcp_data(buf,0,strbuf);
						break;
					case ETHLCD_GET_BUTTONS:
						strbuf[0] = ETHLCD_GET_BUTTONS;
						strbuf[1] = (PINC & 0x3E) | ((PINB & 1<<1) >> 1);
						strbuf[2] = 0;
						plen=fill_tcp_data(buf,0,strbuf);
						break;
					case ETHLCD_SET_BACKLIGHT:	//set the backlight
						if (buf[dat_p+1]==ETHLCD_BACKLIGHT_ON)
						{
							DDRB |= (1<<DDB0);	//set as output
							PORTB &= ~(1<<PB0);	//gnd - full backlight
						}
						else if (buf[dat_p+1]==ETHLCD_BACKLIGHT_HALF)
						{
						        DDRB |= (1<<DDB0);	//set as output
							PORTB |= (1<<PB0);	//vcc - half backlight
						}
						else if (buf[dat_p+1]==ETHLCD_BACKLIGHT_OFF)
						{
							//when set as input - no backlight at all
							//i know this is a dirty hack - but it works :)
							//if you're avr guru then let me know if this
							//is terribly wrong and i have to cut this code out
							//or if i can leave as it is
							DDRB &= ~(1<<DDB0);	//set as input
							PORTB &= ~(1<<PB0);	//gnd
						}
						strbuf[0] = ETHLCD_SET_BACKLIGHT;
						strbuf[1] = 0;
						plen=fill_tcp_data(buf,0,strbuf);
						break;
					case ETHLCD_SET_BEEP:
						if (buf[dat_p+1]==0)
							PORTC &= ~(1<<PC0);	//beep off
						else
							PORTC |= (1<<PC0);	//beep on
						strbuf[0] = ETHLCD_SET_BEEP;
						strbuf[1] = 0;
						plen=fill_tcp_data(buf,0,strbuf);
						break;
					case ETHLCD_GET_FIRMWARE_VERSION:	//firmware version
						plen=fill_tcp_data_p(buf,0,PSTR(FIRMWARE_VERSION));
						break;
					case ETHLCD_GET_PROTOCOL_VERSION:	//protocol version
						strbuf[0] = ETHLCD_GET_PROTOCOL_VERSION;
						strbuf[1] = PROTOCOL_VERSION;
						strbuf[2] = 0;
						plen=fill_tcp_data(buf,0,strbuf);
						break;
					case ETHLCD_GET_ENC_REVISION:	//enc revision
						strbuf[0] = ETHLCD_GET_ENC_REVISION;
						strbuf[1] = enc28j60getrev();
						strbuf[2] = 0;
						plen=fill_tcp_data(buf,0,strbuf);
						break;
					case ETHLCD_CLOSE_CONN:
						strbuf[0] = ETHLCD_CLOSE_CONN;
						strbuf[1] = 0;
						plen=fill_tcp_data(buf,0,strbuf);
						break;
					default:
						strbuf[0] = ETHLCD_UNRECOGNIZED_COMMAND;
						strbuf[1] = 0;
						plen=fill_tcp_data(buf,0,strbuf);
						break;
                                }
                                make_tcp_ack_from_any(buf, mydataport); // send ack
                                make_tcp_ack_with_data(buf,plen, (buf[dat_p] == ETHLCD_CLOSE_CONN) ? 1 : 0); // send data
                                continue;
                        }
                }
                // ethlcd tcp data port end
                //
                // udp is not supported
        }
        return (0);
}
